/*
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * Jini(TM) Technology Core Platform, v.1.0 (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of
 * the License at http://java.sun.com/products/jini. Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * CopyrightVersion v1.0_Jini
 */
package com.sun.jini.thread;
import com.sun.jini.constants.TimeConstants;

import java.text.DateFormat;
import java.util.SortedMap;
import java.io.PrintWriter;

/**
 * A Queue of timed tasks.  Each task implements <code>Runnable</code>.
 * Events can either be executed in the queue's thread or in their own thread.
 * <p>
 * A task is an object that implements <code>Runnable</code>.  It is
 * scheduled by invoking <code>schedule</code> with a time at which it
 * should be run.  When that time arrives (approximately) the task will
 * be pulled off the queue and have its <code>run</code> method invoked.
 * <p>
 * A <code>schedule</code> request can specify a
 * <code>WakeupManager.ThreadDesc</code>, which will define the parameters
 * of a thread to be created to run the <code>Runnable</code>.  You can
 * specify the group, whether the thread is a daemon thread, and the priority.
 * <p>
 * When a tasks is scheduled a <code>WakeupManager.Ticket</code> is returned
 * that can be used to cancel the event if desired.
 * <p>
 * The queue requires its own thread, whose parameters can be defined
 * via a <code>ThreadDesc</code> if desired.  This thread will
 * continue indefinitely unless you invoke <code>stopWhenEmpty</code>
 * with <code>true</code>, which causes the thread to exit when the
 * queue becomes empty.
 *
 * @author Ken Arnold
 *
 * @see java.lang.Runnable
 */
public class WakeupManager implements TimeConstants {
    private SortedMap	contents;	// the queue
    private long        nextBreaker = 0;// Guarded by contents
    private Ticket      head = null;    // First item in contents 
    private boolean	stopWhenEmpty;	// exit when nothing left to do?
    private Kicker	kicker;		// the kicker object
    private ThreadDesc	kickerDesc;	// the desc for kicker threads
    private Thread	kickerThread;	// the thread running the kicker

    private static DateFormat dateFmt  = 
	DateFormat.getTimeInstance(DateFormat.LONG);

    private final static PrintWriter DEBUG =
	ThreadDebug.getWriter("wakeupQueue");

    /**
     * Description of a future thread.
     *
     * @see WakeupManager#schedule
     * @see WakeupManager#WakeupManager(WakeupManager.ThreadDesc)
     */
    public static class ThreadDesc {
	private ThreadGroup group;		// group to create in
	private boolean daemon;			// create as daemon?
	private int priority;			// priority

	/**
	 * Equivalent to
	 * <pre>
	 *     ThreadDesc(null, false)
	 * </pre>
	 */
	public ThreadDesc() {
	    this(null, false);
	}

	/**
	 * Equivalent to
	 * <pre>
	 *     ThreadDesc(group, deamon, Thread.NORM_PRIORITY)
	 * </pre>
	 */
	public ThreadDesc(ThreadGroup group, boolean daemon) {
	    this(group, daemon, Thread.NORM_PRIORITY);
	}

	/**
	 * Describe a future thread that will be created in the given group,
	 * deamon status, and priority.
	 *
	 * @param group The group to be created in.  If <code>null</code>,
	 *		the thread will be created in the default group.
	 * @param daemon The thread will be a daemon thread if this is
	 *		<code>true</code>.
	 * @param priority The thread's priority.
	 */
	public ThreadDesc(ThreadGroup group, boolean daemon, int priority) {
	    this.group = group;
	    this.daemon = daemon;
	    this.priority = priority;
	}

	/**
	 * Create a thread for the given runnable based on the values in this
	 * object.
	 */
	private Thread thread(Runnable r) {
	    Thread thr;
	    if (group == null)
		thr = new Thread(r);
	    else
		thr = new Thread(group, r);
	    thr.setDaemon(daemon);
	    thr.setPriority(priority);
	    return thr;
	}

	public String toString() {
	    return "[" + group + ", " + daemon + ", " + priority + "]";
	}
    }

    /**
     * A ticket that can be used for cancelling a future task.  It describes
     * the task itself as well.
     */
    public static class Ticket implements Comparable {
	/** When the task should occur. */
	public final long when;
	/** The task object to be executed */
	public final Runnable task;
	/** The <code>ThreadDesc</code>, or <code>null</code> if none. */
	public final ThreadDesc desc;

	/** Tie beaker used when two tickets have the same value for when */
	private final long breaker;

        Ticket(long when, Runnable task, ThreadDesc threadDesc,
	     long breaker)
	{
	    if (task == null)
		throw new NullPointerException("task not specified");
	    this.when = when;
	    this.task = task;
	    this.desc = threadDesc;
	    this.breaker = breaker;
	}

	public String toString() {
	    return dateFmt.format(new Long(when)) + "(" + when + ")" + ", "
		+ task.getClass().getName() + ", " + desc;
	}

	public boolean equals(Object o) {
	    if (!(o instanceof Ticket))
		return false;

	    final Ticket that = (Ticket)o;

	    return that.when == when && that.breaker == breaker;
	}

	public int compareTo(Object o) {
	    final Ticket that = (Ticket)o;	    
	    
	    final long whenDiff = when - that.when;
	    if (whenDiff > 0)
		return 1;
	    else if (whenDiff < 0)
		return -1;
	    else {
		final long breakerDiff = breaker - that.breaker;	

		if (breakerDiff > 0)
		    return 1;
		else if (breakerDiff < 0)
		    return -1;
		else
		    return 0;
	    }
	}
    }

    /**
     * Create a new <code>WakeupManager</code>. Equivalent to.
     * <pre>
     *     WakeupManager(new ThreadDesc())
     * </pre>
     *
     * @see WakeupManager.ThreadDesc
     */
    public WakeupManager() {
	this(new ThreadDesc());
    }

    /**
     * Create a new <code>WakeupManager</code>.  The thread used for
     * timing will be created according to the provided <code>ThreadDesc</code>.
     */
    public WakeupManager(ThreadDesc desc) {
	kickerDesc = desc;
	stopWhenEmpty = false;
	contents = new java.util.TreeMap();
	kicker = new Kicker();
	kickerThread = kickerDesc.thread(kicker);
	kickerThread.start();
    }

    /**
     * Schedule the given task for the given time.  The task's <code>run</code>
     * method will be executed synchronously in the queue's own thread, so it
     * should be brief or it will affect whether future events will be executed
     * at an appropriate time.
     */
    public Ticket schedule(long when, Runnable task) {
	return schedule(when, task, null);
    }

    /**
     * Schedule the given task for the given time, to be run in a thread.
     * When the time comes, a new thread will be created according to the
     * <code>ThreadDesc</code> object provided.  If <code>threadDesc</code> is
     * <code>null</code>, this is equivalent to the other form of
     * <code>schedule</code>.
     */
    public Ticket schedule(long when, Runnable task, ThreadDesc threadDesc) {
	if (task == null)
	    throw new NullPointerException("task");

	synchronized (contents) {
	    Ticket t = new Ticket(when, task, threadDesc, nextBreaker++);
	    contents.put(t,t);
	    checkHead();
	    return t;
	}
    }

    /**
     * Cancel the given ticket.
     */
    public void cancel(Ticket t) {
	synchronized (contents) {
	    contents.remove(t);
	    checkHead();
	}
    }

    /**
     * Cancel all tickets.
     */
    public void cancelAll() {
	synchronized (contents) {
	    contents.clear();
	    checkHead();
	}
    }


    /** Called whenever we change contents to see if there is a new value
	for head to see if we have to tell the kicker */
    private void checkHead() {
        // We are always called from code syncronized on contents
	final Ticket oldHead = head;

	if (contents.isEmpty())
	    head = null;
	else
	    head = (Ticket)contents.firstKey();

	if (head == oldHead) return;

	// New first event, kicker needs to wake up and change its sleep time
	kicker.newTime();
    }

    /**
     * Return whether this queue stops executing when it's empty.
     */
    public boolean stopWhenEmpty() {
	return stopWhenEmpty;
    }

    /**
     * Set whether this queue stops executing when it's empty.
     */
    public void stopWhenEmpty(boolean stop) {
	stopWhenEmpty = stop;
    }

    /**
     * Return whether the queue is currently empty.
     */
    public boolean isEmpty() {
	synchronized (contents) {
	    return (contents.isEmpty());
	}
    }

    /**
     * Stop executing.
     */
    public void stop() {
	kicker.kill();
    }

    /**
     * The kicker work.  This is what sleeps until the time of
     * the next event.
     */
    private class Kicker implements Runnable {
	private long sleepTime;		// time to sleep until the next
	private boolean die = false;	// should I die instead of continue?

	/**
	 * Keep running along, executing tasks and sleeping.  Most of
	 * the work is in the <code>doTasks</code> method.
	 */
	public void run() {
	    while (!die) {
		doTasks();		// figures out the next sleepTime
		if (stopWhenEmpty && sleepTime == Long.MAX_VALUE)
		    return;

		    if (DEBUG != null) {
			DEBUG.println("sleepTime = " +
			    (sleepTime == Long.MAX_VALUE ? "***FOREVER***" :
							   sleepTime + ""));
		    }

		    synchronized(this) {
		        try {
		            if (sleepTime > 0)	// interruption might cause this
			        wait(sleepTime);
		        } catch (InterruptedException e) {
		        }
		    }
	    }
	}

	/**
	 * Kill off this thread
	 */
	private void kill() {
	    die = true;
	    kickerThread.interrupt();
	}

	/**
	 * A new time has arrived -- if we're sleeping, wake up.
	 */
	private synchronized void newTime() {
	        notifyAll();
	}

	/**
	 * Execute the tasks whose time has come, and calculate the
	 * next sleep time.
	 */
	private void doTasks() {
	    Ticket t;

	    do {
		synchronized (contents) {
		    if (contents.isEmpty()) {           // nothing to do
			sleepTime = Long.MAX_VALUE;	// sleep forever
			continue;
		    }

		    t = (Ticket) contents.firstKey();
		}

		sleepTime = t.when - System.currentTimeMillis();
		if (sleepTime <= 0) {		// it's time has come

		    // consume it
		    synchronized (contents) { 
			Ticket t2 = (Ticket)contents.remove(t);
			if (t2 != t)
			    // Someone must have not wanted it to
			    // run and removed it! 
                            continue;
		    }

		    // run it
		    if (t.desc == null)
			t.task.run();		// ... in this thread
		    else			// ... in its own thread
			t.desc.thread(t.task).start();

		    // we may have been interrupted to die
		    if (kickerThread.isInterrupted())
			return;
		}
	    // While the head's when may already be here...
	    } while (sleepTime < 0);
	}
    }
}
